import hashlib

from bifas.utils.data_structure.binary import Binary

def sha256(
    x:Binary=Binary(x=0),
) -> Binary:
    """
    Using SHA256 to hash Binary to fixed length of 256 bits Binary.

    Parameters
    ----------
    x : Binary, default Binary(x=0)

    Returns
    -------
    Binary
        The hash value stored as Binary.

    See Also
    --------
    bifas.utils.data_structure.binary.Binary

    Examples
    --------
    >>> from bifas.utils.data_structure.binary import Binary
    >>> from bifas.utils.cryptography.hash import sha256
    >>> question = b"Blockchain International Financial Assets System (BIFAS)"
    >>> answer = "55affbc3b2c2b989f32e3b6cae1ca4bd154a7347adcf380297b484e9d387f5c0"
    >>> answer == sha256(Binary(x=question)).get_x(data_type="str")
    True
    """
    return Binary(
        hashlib.sha256(
            x.get_x(data_type="bytes")
        ).digest()
    )


import typing
import math

from bifas.utils.data_structure.date_time import current_date_time_str

def int_to_bytes(
    x:int = 0,
    endianness:typing.Literal["big","little"]="big",
    signed_flag=False,
) -> bytes:
    if (x == 0):
        return b"\x00"
    return x.to_bytes(
        length=math.ceil(
            math.log(
                x + 1,  # prevent  boundary case of 256, 65536, ...
                2 ** 8,
            ),
        ),
        byteorder=endianness,
        signed=signed_flag,
    )

class Binary():
    """
    Generic class for handling binary data.

    See Also
    --------
    bifas.utils.cryptography.hash.sha256
    """
    def __init__(
        self,
        x:typing.Union[
            bytes,
            str,
            int,
        ],
        endianness:typing.Literal["big","little"]="big",
        signed_flag=False,
    ) -> None:
        self.__x = None
        self.endianness = endianness
        self.signed_flag = signed_flag
        if (type(x) == bytes):
            self.__x = x
        elif (type(x) == str):
            x = int(x, 16)
            self.__x = int_to_bytes(x)
        elif (type(x) == int):
            self.__x = int_to_bytes(x)
        else:
            raise Exception(
                "{} : [CRITICAL] bifas.utils.data_structure.binary.Binary.__init__() unknown type of x = {}".format(
                    current_date_time_str(),
                    x,
                )
            )
    
    def get_x(
        self,
        data_type:typing.Literal[
            "bytes",
            "str",
            "int",
        ]="bytes"
    ) -> typing.Union[
        bytes,
        str,
        int,
    ]:
        """
        Generic function to return the value in various format.

        Parameters
        ----------
        data_type : typing.Literal[
            "bytes",
            "str",
            "int",
        ], default "bytes"

        Returns
        -------
        typing.Union[
            bytes,
            str,
            int,
        ]
            The value with respective data type.

        See Also
        --------

        Examples
        --------
        >>> from utils.data_structure.binary import Binary
        >>> Binary(b"\x6b").get_x(data_type="bytes") == b"\x6b"
        True
        >>> Binary(b"\x04\x5e").get_x(data_type="bytes") == b"\x04\x5e"
        True
        >>> Binary(b"\x6b").get_x(data_type="str") == "6b"
        True
        >>> Binary(b"\x04\x5e").get_x(data_type="str") == "45e"
        True
        >>> Binary(b"\x6b").get_x(data_type="int") == 107
        True
        >>> Binary(b"\x04\x5e").get_x(data_type="int") == 1118
        True
        """
        if (data_type == "bytes"):
            return self.__x
        elif (data_type == "str"):
            return format(int.from_bytes(bytes=self.__x,byteorder=self.endianness, signed=self.signed_flag), "x")
        elif (data_type == "int"):
            return int.from_bytes(bytes=self.__x,byteorder=self.endianness, signed=self.signed_flag)
        else:
            raise Exception(
                "{} : [CRITICAL] bifas.utils.data_structure.binary.Binary.get_x() unknown data_type = {}".format(
                    current_date_time_str(),
                    data_type,
                )
            )
    
    def __add__(self, x2) -> bytes:
        """
        XOR logical operation on two binary number.

        This function simply wraps the ``+`` operator to perform XOR logical operation on two Binary class objects.

        Parameters
        ----------
        self :
            Operand A.
        x2 : Binary
            Operand B.

        Returns
        -------
        bytes
            The XOR of ``self`` and ``x2``.

        See Also
        --------

        Examples
        --------
        >>> from utils.data_structure.binary import Binary
        >>> Binary(b"\x6b") + Binary(b"\x04\x5e") == b"\x04\x35"
        True
        """
        return int_to_bytes(self.get_x(data_type="int") ^ x2.get_x(data_type="int"))


# Stateless function of generating random large prime given the bit-length, return as utils.data_structure.binary.Binary.
# Stateless function of generating private key from two large prime numbers, return as utils.data_structure.binary.Binary.
# Stateless function of generating public key from two large prime numbers, return as utils.data_structure.binary.Binary.
# Stateless function of digital signature or encrypt from private key or public key on the message, input as utils.data_structure.binary.Binary, return as utils.data_structure.binary.Binary.
# Stateless boolean function to verify digital signed message (utils.data_structure.binary.Binary) with public key or private key.


# Stateless function of generating private key from arbitrary passphrase, return as utils.data_structure.binary.Binary.
# Stateless function of generating public key from private key, return as utils.data_structure.binary.Binary.
# Stateless function of generating account address from public key, return as utils.data_structure.binary.Binary.
# Stateless function of digital signature from private key and message (utils.data_structure.binary.Binary), return as utils.data_structure.binary.Binary.
# Stateless boolean function to verify digital signed message (utils.data_structure.binary.Binary) with public key (utils.data_structure.binary.Binary).


# python3 -m unittest bifas.utils.cryptography.test_hash

import unittest

from bifas.utils.data_structure.binary import Binary
from bifas.utils.cryptography.hash import sha256

class TestBinary(unittest.TestCase):

    def test_hard_code_case(self):
        question = b"Blockchain International Financial Assets System (BIFAS)"
        answer = "55affbc3b2c2b989f32e3b6cae1ca4bd154a7347adcf380297b484e9d387f5c0"
        self.assertEqual(
            answer,
            sha256(Binary(x=question)).get_x(data_type="str"),
        )

if __name__ == "__main__":
    unittest.main()


# python3 -m unittest bifas.utils.data_structure.test_binary

import unittest
import random

from bifas.utils.data_structure.binary import Binary, int_to_bytes
from bifas.utils.data_structure.date_time import current_date_time_str

NUM_OF_BYTE = 2
NUM_OF_Trials = 10 ** 4

class TestBinary(unittest.TestCase):

    def test_hard_code_case(self):
        __tmp_a = {
            "bytes_value": b"\x6b",
            "str_value": "6b",
            "int_value": 107,
        }
        __tmp_b = {
            "bytes_value": b"\x04\x5e",
            "str_value": "45e",
            "int_value": 1118,
        }
        __obj_a = Binary(__tmp_a["int_value"])
        __obj_b = Binary(__tmp_b["int_value"])
        
        self.assertEqual(
            b"\x04\x35",
            __obj_a + __obj_b,
        )
        self.assertEqual(Binary(b"\x6b") + Binary(b"\x04\x5e"), b"\x04\x35")

        self.assertEqual(Binary(b"\x6b").get_x(data_type="bytes"), b"\x6b")   
        self.assertEqual(Binary(b"\x04\x5e").get_x(data_type="bytes"), b"\x04\x5e")
        self.assertEqual(Binary(b"\x6b").get_x(data_type="str"), "6b")   
        self.assertEqual(Binary(b"\x04\x5e").get_x(data_type="str"), "45e")
        self.assertEqual(Binary(b"\x6b").get_x(data_type="int"), 107)   
        self.assertEqual(Binary(b"\x04\x5e").get_x(data_type="int"), 1118)

    def __gen_int_str_bytes_test_case(self) -> dict:
        answer = {
            "int_value" : random.randint(
                0,
                2**(NUM_OF_BYTE * 8) - 1,
            ),
        }
        answer["str_value"] = format(answer["int_value"], "x")
        answer["bytes_value"] = int_to_bytes(answer["int_value"])
        return answer


    def test_init_with_binary(self):
        for i in range(NUM_OF_Trials):
            __tmp = self.__gen_int_str_bytes_test_case()
            print(
                "{} : [INFO] utils.data_structure.test_binary.TestBinary.test_init_with_binary() i={} int_value= {} str_value= {} bytes_value= {}".format(
                    current_date_time_str(),
                    i,
                    __tmp["int_value"],
                    __tmp["str_value"],
                    __tmp["bytes_value"],
                ),
            )
            __obj = Binary(__tmp["bytes_value"])
            self.assertEqual(__obj.get_x(data_type="bytes"), __tmp["bytes_value"])
            self.assertEqual(__obj.get_x(data_type="str"), __tmp["str_value"])
            self.assertEqual(__obj.get_x(data_type="int"), __tmp["int_value"])
    
    def test_init_with_str(self):
        for i in range(NUM_OF_Trials):
            __tmp = self.__gen_int_str_bytes_test_case()
            print(
                "{} : [INFO] utils.data_structure.test_binary.TestBinary.test_init_with_str() i={} int_value= {} str_value= {} bytes_value= {}".format(
                    current_date_time_str(),
                    i,
                    __tmp["int_value"],
                    __tmp["str_value"],
                    __tmp["bytes_value"],
                ),
            )
            __obj = Binary(__tmp["str_value"])
            self.assertEqual(__obj.get_x(data_type="bytes"), __tmp["bytes_value"])
            self.assertEqual(__obj.get_x(data_type="str"), __tmp["str_value"])
            self.assertEqual(__obj.get_x(data_type="int"), __tmp["int_value"])
    
    def test_init_with_int(self):
        for i in range(NUM_OF_Trials):
            __tmp = self.__gen_int_str_bytes_test_case()
            print(
                "{} : [INFO] utils.data_structure.test_binary.TestBinary.test_init_with_int() i={} int_value= {} str_value= {} bytes_value= {}".format(
                    current_date_time_str(),
                    i,
                    __tmp["int_value"],
                    __tmp["str_value"],
                    __tmp["bytes_value"],
                ),
            )
            __obj = Binary(__tmp["int_value"])
            self.assertEqual(__obj.get_x(data_type="bytes"), __tmp["bytes_value"])
            self.assertEqual(__obj.get_x(data_type="str"), __tmp["str_value"])
            self.assertEqual(__obj.get_x(data_type="int"), __tmp["int_value"])
    
    def test_xor(self):
        for i in range(NUM_OF_Trials):
            __tmp_a = self.__gen_int_str_bytes_test_case()
            __tmp_b = self.__gen_int_str_bytes_test_case()
            print(
                "{} : [INFO] utils.data_structure.test_binary.TestBinary.test_xor() i={} int_value= {}, {} str_value= {}, {} bytes_value= {}, {}".format(
                    current_date_time_str(),
                    i,
                    __tmp_a["int_value"],
                    __tmp_b["int_value"],
                    __tmp_a["str_value"],
                    __tmp_b["str_value"],
                    __tmp_a["bytes_value"],
                    __tmp_b["bytes_value"],
                ),
            )
            __obj_a = Binary(__tmp_a["int_value"])
            __obj_b = Binary(__tmp_b["int_value"])
            
            self.assertEqual(
                int_to_bytes(__obj_a.get_x(data_type="int") ^ __obj_b.get_x(data_type="int")),
                __obj_a + __obj_b,
            )

if __name__ == "__main__":
    unittest.main()